////////////////////////////////////////////////////////////////////////
//
// FeatureVertexFinder class
//
// jasaadi@fnal.gov
//
// This algorithm is designed to reconstruct the vertices using the
// 2D cluster information and CornerFinderAlg to find 2-d and 3-d verticies
//
// We will utilize many of the methods found in VeretxFinder2d but modify them 
// to return vertex candidates and only a point that has matching in at least 
// 2 views will be used
//
//
//
// This is Preliminary Work and needs modifications
//
// ////////////////////////////////////////////////////////////////////////

// ##########################
// ### Basic C++ Includes ###
// ##########################
#include <vector>
#include <string>
#include <iostream>
#include <iomanip>
#include <ios>
#include <sstream>
#include <fstream>
#include <math.h>
#include <algorithm>
#include <vector>

// ##########################
// ### Framework Includes ###
// ##########################
#include "art/Framework/Core/ModuleMacros.h" 
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Principal/Event.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// ########################
// ### LArSoft Includes ###
// ########################
#include "RecoBase/EndPoint2D.h"
#include "RecoBase/Hit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Track.h"
#include "RecoBase/Shower.h"
#include "RecoBase/Vertex.h"
#include "Geometry/Geometry.h"
#include "Geometry/CryostatGeo.h"
#include "Geometry/TPCGeo.h"
#include "Geometry/PlaneGeo.h"
#include "Geometry/WireGeo.h"
#include "Utilities/LArProperties.h"
#include "Utilities/DetectorProperties.h"
#include "Utilities/AssociationUtil.h"
#include "RecoAlg/CornerFinderAlg.h"
#include "RecoAlg/ClusterParamsAlg.h"

// #####################
// ### ROOT Includes ###
// #####################
#include "TMath.h"
#include "TH1D.h"
#include "TVectorD.h"
#include "TGeoManager.h"
#include "TMath.h"
#include "TGraph.h"
#include "TF1.h"
#include "TVector3.h"

// =====================================================================================================
// =====================================================================================================
#ifndef FeatureVertexFinder_H
#define FeatureVertexFinder_H

class TH1D;

///vertex reconstruction
namespace vertex {
   
 class FeatureVertexFinder :  public art::EDProducer {
    
  public:
    
    explicit FeatureVertexFinder(fhicl::ParameterSet const& pset); 
    virtual ~FeatureVertexFinder();        
    void beginJob();
    void reconfigure(fhicl::ParameterSet const& p);

    
    void produce(art::Event& evt);

  private:
    
    TH1D *dtIC;
  
    std::string fClusterModuleLabel;
    std::string fHitModuleLabel;
    // JA:Adding Corner Finder Alg
    cluster::CornerFinderAlg  fCorner;
    cluster::ClusterParamsAlg fClParAlg;
  };
    
}//<---End namespace vertex

#endif // FeatureVertexFinder_H
// =====================================================================================================
// =====================================================================================================








// =====================================================================================================
// =====================================================================================================
namespace vertex{

//-----------------------------------------------------------------------------
  FeatureVertexFinder::FeatureVertexFinder(fhicl::ParameterSet const& pset):
  fCorner(pset.get<fhicl::ParameterSet>("CornerPset")),
  fClParAlg(pset.get<fhicl::ParameterSet>("ClusterParamsAlg"), pset.get< std::string >("module_type"))
  {  
    /*this->*/reconfigure(pset);    
    produces< std::vector<recob::Vertex> >();
    produces< std::vector<recob::EndPoint2D> >();
    produces< art::Assns<recob::EndPoint2D, recob::Hit> >();
    produces< art::Assns<recob::Vertex, recob::Hit> >();
    produces< art::Assns<recob::Vertex, recob::Shower> >();
    produces< art::Assns<recob::Vertex, recob::Track> >();
  }
//-----------------------------------------------------------------------------
  FeatureVertexFinder::~FeatureVertexFinder()
  {
  }

  //---------------------------------------------------------------------------
  void FeatureVertexFinder::reconfigure(fhicl::ParameterSet const& p) 
  {
    fClusterModuleLabel  = p.get< std::string >("ClusterModuleLabel");
    fHitModuleLabel	 = p.get< std::string >("HitModuleLabel");
    return;
  }
  //-------------------------------------------------------------------------
  void FeatureVertexFinder::beginJob(){
    // get access to the TFile service
    art::ServiceHandle<art::TFileService> tfs;
    dtIC = tfs->make<TH1D>("dtIC","It0-Ct0",100,-5,5);
    dtIC->Sumw2();
  }

// //-----------------------------------------------------------------------------
  void FeatureVertexFinder::produce(art::Event& evt)
  {
  
  
    //std::cout<<"Top of FeatureVertexFinder Produce"<<std::endl;  
    // #########################
    // ### Geometry Services ###
    // #########################
    art::ServiceHandle<geo::Geometry> geom;
    
    // ###############################
    // ### LAr Properties Services ###
    // ###############################
    art::ServiceHandle<util::LArProperties> larprop;
    
    // ####################################
    // ### Detector Properties Services ###
    // ####################################
    art::ServiceHandle<util::DetectorProperties> detprop;
    
    // ########################################
    // ### cout Run Number and Event Number ###
    // ########################################
    std::cout<< std::endl;
    std::cout << " ======================================================" << std::endl;
    std::cout << "Run    : " << evt.run() <<" Event  : "<<evt.id().event() << std::endl;
    std::cout << " ======================================================" << std::endl;
    std::cout << std::endl;

    
    
    // ###############################
    // ### Defining TPC Parameters ###
    // ###############################
    // === TPC Name === (JA: Not sure if I need this)
    TString tpcName = geom->GetLArTPCVolumeName();
    // === Length of the detector ===
    double DetectorLength =  (geom->DetHalfHeight())*2.;
    // === wire angle with respect to the vertical direction ===
    double Angle = geom->Plane(1).Wire(0).ThetaZ(false)-TMath::Pi()/2.; 
    
    // ==================================================
    // === Calculating the Timetick to CM conversion  ===
    // ==================================================
    double TimeTick = detprop->SamplingRate()/1000.; //<---To get units of microsecond...not nanosec
    double Efield_drift = larprop->Efield();      // Electric Field in the drift region in kV/cm
    double Temperature  = larprop->Temperature(); // LAr Temperature in K
    // === Drift velocity in the drift region (cm/us) ===
    double DriftVelocity = larprop->DriftVelocity(Efield_drift,Temperature); 
    // === Time sample (cm) === 
    double TimetoCm = DriftVelocity*TimeTick; 
    
    
    double presamplings = detprop->TriggerOffset(); //trigger offset
    
    // =============================================
    // === Wire Pitch variable (to be set later) ===
    // =============================================
    double wire_pitch   = 0;
    // ===========================================================
    // === Number of planes in this detector (to be set later) ===
    // ===========================================================
    int nplanes = 0;
    
    std::vector<double> vtx_wire = {0.};
    std::vector<double> vtx_time = {0.};
    std::vector<double> vtx_plane = {0.};
    
    
    
    
    
//-------------------------------------------------------------------------------------------------------------------------------------------------   
    // ####################################################################
    // ### Utilizing the CornerFinderAlg to get a list of EndPoint2d's  ###
    // ####################################################################
    
    std::vector<double>   feature_wire = {0.};
    std::vector<double>   feature_time = {0.};
    std::vector<double>   feature_plane = {0.};
    std::vector<double>	  feature_strength = {0.};
    std::vector<uint32_t> feature_channel =  {0};
    int nFeatures = 0;
    
    float x_feature[1000] = {0.}, y_feature[1000] = {0.}, z_feature[1000] = {0.}, strength_feature[1000] = {0.};
    double yy = 0., zz = 0.;
    int n3dFeatures = 0;
    
    
/*    
    // ###################################################
    // ### Take in the raw information about the event ###
    // ###################################################
    fCorner.TakeInRaw(evt);
    
    // #######################################################
    // ### Push the features into a vector of Endpoint2d's ###
    // #######################################################
    std::vector<recob::EndPoint2D> EndPoints;
    fCorner.get_feature_points(EndPoints);
    
    // ########################################################
    // ### Loop over all the features and record their info ###
    // ########################################################
    for(size_t i=0; i!=EndPoints.size(); ++i)
    	{
	
	feature_wire.push_back(EndPoints.at(i).WireID().Wire);
	feature_time.push_back(EndPoints.at(i).DriftTime());
	feature_plane.push_back(EndPoints.at(i).WireID().Plane);
	feature_channel.push_back(geom->PlaneWireToChannel(EndPoints.at(i).WireID().Plane, EndPoints.at(i).WireID().Wire, 0, 0));
	feature_strength.push_back(EndPoints.at(i).Strength());
	nFeatures++;
	
	//std::cout<<std::endl;
	//std::cout<<"EndPoints View() = "<<EndPoints.at(i).WireID().Plane<<std::endl;
	//std::cout<<"EndPoints DriftTime() = "<<EndPoints.at(i).DriftTime()<<std::endl;
	//std::cout<<"EndPoints Wire = "<<EndPoints.at(i).WireID().Wire<<std::endl;
	//std::cout<<"EndPoints Channel = "<<geom->PlaneWireToChannel(EndPoints.at(i).WireID().Plane, EndPoints.at(i).WireID().Wire, 0, 0)<<std::endl;
	//std::cout<<std::endl;
	
	
       	}//<---End i loop finding 2d Features
    
    
    
    // ##############################
    // ### Looping over cryostats ###
    // ##############################
    for(size_t cstat = 0; cstat < geom->Ncryostats(); ++cstat)
    	{
    	// ##########################
      	// ### Looping over TPC's ###
      	// ##########################
      	for(size_t tpc = 0; tpc < geom->Cryostat(cstat).NTPC(); ++tpc)
		{
     
    		// ##############################################################################
    		// ### Now loop over those features and see if they match up in between views ###
    		// ##############################################################################
    		for(int feature1 = nFeatures; feature1 > 0; feature1 --)
    			{
			for(int feature2 = 0; feature2 < feature1; feature2++)
				{
				// ##########################################################################
				// ### Check to make sure we are comparing features from different planes ###
				// ##########################################################################
				if(feature_plane[feature1] != feature_plane[feature2])
					{
					
					// #############################################################################
					// ### Get the appropriate time offset for the two planes we are considering ###
					// #############################################################################
					float tempXFeature1 = detprop->ConvertTicksToX(feature_time[feature1], feature_plane[feature1], tpc, cstat);
					
					float tempXFeature2 = detprop->ConvertTicksToX(feature_time[feature2], feature_plane[feature2], tpc, cstat);
					
					
					// ####################################################################
					// ### Checking to see if these features have intersecting channels ###
					// ###               and are within 1 cm in projected X             ###
					// ####################################################################
					if( geom->ChannelsIntersect( feature_channel[feature2], feature_channel[feature1], yy, zz) &&
					    std::abs(tempXFeature1 - tempXFeature2) < 0.5)
					    	{
				
						x_feature[n3dFeatures] = detprop->ConvertTicksToX(feature_time[feature1], feature_plane[feature1], tpc, cstat); 
						y_feature[n3dFeatures] = yy;
						z_feature[n3dFeatures] = zz;
						strength_feature[n3dFeatures] = feature_strength[feature1] + feature_strength[feature1];
				
				
						//std::cout<<std::endl;
						//std::cout<<"feature_time[feature1] = "<<feature_time[feature1]<<std::endl;
						//std::cout<<"### Found a 3d Feature ###"<<std::endl;
						//std::cout<<"X = "<<x_feature[n3dFeatures]<<" , Y = "<<y_feature[n3dFeatures]<<" , Z = "<<z_feature[n3dFeatures]<<std::endl;
						//std::cout<<"Strength = "<<strength_feature[n3dFeatures]<<std::endl;
						//std::cout<<std::endl;
						n3dFeatures++;
	
						}//<---End Checking if the feature matches in 3d
					}//<---End Checking features are in different planes
				}//<---End feature2 loop
			}//<---End feature1 loop
		}//<---End looping over TPC's
	}//<---End looping over cryostats
 	
*/	
 
 
 
 
 
//-------------------------------------------------------------------------------------------------------------------------------------------------    
    
    //std::cout<<"Making new collections for output"<<std::endl;
    //Point to a collection of vertices to output.
    std::unique_ptr<std::vector<recob::Vertex> >                 vcol(new std::vector<recob::Vertex>);          //3D vertex
    std::unique_ptr<std::vector<recob::EndPoint2D> >             epcol(new std::vector<recob::EndPoint2D>);  //2D vertex
    std::unique_ptr< art::Assns<recob::EndPoint2D, recob::Hit> > assnep(new art::Assns<recob::EndPoint2D, recob::Hit>);
    std::unique_ptr< art::Assns<recob::Vertex, recob::Shower> >  assnsh(new art::Assns<recob::Vertex, recob::Shower>);
    std::unique_ptr< art::Assns<recob::Vertex, recob::Track> >   assntr(new art::Assns<recob::Vertex, recob::Track>);
    std::unique_ptr< art::Assns<recob::Vertex, recob::Hit> >     assnh(new art::Assns<recob::Vertex, recob::Hit>);
    

    // ###################################################
    // ### Retreiving the Cluster Module for the event ###
    // ###################################################
    art::Handle< std::vector<recob::Cluster> > clusterListHandle;
    evt.getByLabel(fClusterModuleLabel,clusterListHandle);
    
    // #################################################
    // ### Finding hits associated with the clusters ###
    // #################################################
    art::FindManyP<recob::Hit> fmh(clusterListHandle, evt, fClusterModuleLabel);
    
   
    
    // ##################################
    // ### Filling the Cluster Vector ###
    // ##################################
    art::PtrVector<recob::Cluster> clusters;
    for (unsigned int ii = 0; ii <  clusterListHandle->size(); ++ii){
      art::Ptr<recob::Cluster> clusterHolder(clusterListHandle,ii);
      clusters.push_back(clusterHolder);
      
    }//<---End ii loop
    
    // ###############################################
    // ### Retreiving the Hit Module for the event ###
    // ###############################################
    art::Handle< std::vector<recob::Hit> > hitListHandle;
    evt.getByLabel(fHitModuleLabel,hitListHandle);
    
    // #########################################
    // ### Putting Hits into a vector (hits) ###
    // #########################################
    std::vector< art::Ptr<recob::Hit> > hits;
    art::fill_ptr_vector(hits, hitListHandle);
    
    



// -------------------------------------------------------------------------------------------------------------------------

// #######################################################################
// ###   Trying to incorporate ClusterParmsAlg into here to utilize    ###
// ### the finding axis of a cluster instead of doing simple pol1 fits ###
// #######################################################################

/*    // ##########################################
    // ### Start by looping over the clusters ###
    // ##########################################
    for(unsigned int iClust = 0; iClust < clusterListHandle->size(); iClust++)
    	{

    	art::Ptr<recob::Cluster> cl(clusterListHandle, iClust);
	
   	std::vector< art::Ptr<recob::Hit> > hitlist = fmh.at(iClust);
	
	double lineslope, lineintercept,goodness,wire_start,time_start,wire_end,time_end;

	fClParAlg.Find2DAxisRough(lineslope,lineintercept,goodness,hitlist);
	// hmmm...seems that my hitlist is always too small....that's strange....
	
	std::cout<<std::endl;
	std::cout<<"iClust = "<<iClust<<" , StartWire = "<<cl->StartPos()[0]<<" , StartTime = "<<cl->StartPos()[1]<<std::endl;
	std::cout<<"iClust = "<<iClust<<" , EndWire   = "<<cl->EndPos()[0]<<" , EndTime   = "<<cl->EndPos()[1]<<std::endl;
	std::cout<<"iClust = "<<iClust<<" , Slope     = "<<lineslope<<" , intercept = "<<lineintercept<<" , goodness = "<<goodness<<std::endl;
	
	
	fClParAlg.Find2DStartPointsHighCharge( hitlist,wire_start,time_start,wire_end,time_end);
	
	std::cout<<std::endl;
	std::cout<<"Recalculated StartWire = "<<wire_start<<" , Recalculated StartTime = "<<time_start<<std::endl;
	std::cout<<"Recalculated EndWire   = "<<wire_end<<" , Recalculated EndTime = "<<time_end<<std::endl;


	}//<---End Cluster list

*/

















// ----------------------------------------------------------------------------------------------------------------------------    
    


    // ==============================================================================
    // ===  Loop over all the clusters that are found in the cluster list and fit ===
    // ==  the hits in those clusters to establish a vector of slopes (dtdwstart) ===
    // === that has the same number and order as our cluster PtrVector (clusters) ===
    // ==============================================================================
    std::vector<int> Cls[3]; //<---- Index to clusters in each view
    std::vector<double> dtdwstart; //<----Slope (delta Time Tick vs delta Wire) 
    std::vector<double> dtdwstartError; //<---Error on the slope    
    
    std::cout<<"cluster.size() = "<<clusters.size()<<std::endl;
    
    //std::cout<<"Getting ready to loop over clusters"<<std::endl;
    for(size_t iclu = 0; iclu < clusters.size(); ++iclu)
    	{
	//std::cout<<"Looping over the first cluster"<<std::endl;
	// ######################################################
	// ### Determine which view the current cluster is in ###
	// ######################################################
	switch(clusters[iclu]->View())
		{
	    	//std::cout<<"Inside the switch loop"<<std::endl;
	  	case geo::kU :
	    	  Cls[0].push_back(iclu);
	          break;
	  	case geo::kV :
	  	  Cls[1].push_back(iclu);
	  	  break;
	  	case geo::kZ :
	  	  Cls[2].push_back(iclu);
	  	  break;
	  	default :
	  	  break;
	  	}
	  
	// #############################################################
	// ### Filling wires and times into a TGraph for the cluster ###
	// #############################################################	  
	std::vector<double> wires;
	std::vector<double> times;
	  
	// ### Gathering the hits associated with the current cluster ###
	std::vector< art::Ptr<recob::Hit> > hit = fmh.at(iclu);
	
	// ### Counting the number of hits in the current cluster (n) ###
	int n = 0;
	// ############################################
	// ### Looping over the hits in the cluster ###
	// ############################################
	//std::cout<<"Getting ready to loop over this clusters hits"<<std::endl;
	for(size_t i = 0; i < hit.size(); ++i)
		{
		// ### JA: Should I convert wire and time to cm? Will that help the fitting? ###
		
	    	wires.push_back(hit[i]->WireID().Wire);
	    	times.push_back(hit[i]->PeakTime());
	    	++n;
	  	}//<---End loop over hits (i)
	// ################################################################
	// ### If there are 2 or more hits in the cluster fill a TGraph ###
	// ###         and fit a from a polynomial or order 1           ###
	// ################################################################
	if(n>=2)
		{
		// ###################################
		// ### Push the hits into a TGraph ###
		// ###################################
	    	TGraph *the2Dtrack = new TGraph(n,&wires[0],&times[0]);  
		// === Try to fit the TGraph with a 1st order polynomial ===
	    	try
			{
	      		the2Dtrack->Fit("pol1","Q");
	      		TF1 *pol1=(TF1*) the2Dtrack->GetFunction("pol1");
	      		double par[2];
			double parerror[2];
	      		pol1->GetParameters(par);
			parerror[1] = pol1->GetParError(1);
			double FitChi2 = pol1->GetChisquare();
			double FitNDF = pol1->GetNDF();
			
			double fitGoodness = FitChi2/FitNDF;
			
			// ############################################################
			// ### Skipping the fitted slope if the 0.5 < Chi2/NDF < 5  ###
			// ############################################################
			
			
			if( fitGoodness > 5)
				{
				mf::LogWarning("FeatureVertexFinder") << "Fitter returned poor Chi2/NDF, using the clusters default dTdW()";
				dtdwstart.push_back(clusters[iclu]->dTdW());
				//std::cout<<"Skipping the fitted slope"<<std::endl;
				//delete the2Dtrack;
				continue;
				
				}//<---End check on chi2/ndf fit
	      		//std::cout<<"Cluster = "<<iclu<<" , Fit Slope = "<<par[1]<<" Clusters Saved Slope = "<<clusters[iclu]->dTdW()<<std::endl;
			//std::cout<<"Cluster Fit Chi2 = "<<FitChi2<<" , Cluster NDF = "<<FitNDF<<" , Chi2/NDF = "<<fitGoodness<<std::endl;
	      
	      		// #######################################################################
	      		// ### Take change in time tick vs change in wire (dT/dW) from the fit ###
	      		// #######################################################################
	      		dtdwstart.push_back(par[1]);
			dtdwstartError.push_back(parerror[1]);
	    		}//<---End try to fit with a polynomial order 1
			
	    	// ############################################################
	    	// ### If the fitter fails just take dT/dW from the cluster ###
	    	// ############################################################
	    	catch(...)
			{
	      		mf::LogWarning("FeatureVertexFinder") << "Fitter failed, using the clusters default dTdW()";
	      		delete the2Dtrack;
	      		dtdwstart.push_back(clusters[iclu]->dTdW());
	      		continue;
	    		}
	    	delete the2Dtrack;
	  	}//<---End if the cluster has 2 or more hits
	// #################################################
	// ### If the cluster has fewer than 2 hits just ### 
	// ###      take the dT/dW from the cluster      ###
	// #################################################
	else {dtdwstart.push_back(clusters[iclu]->dTdW());}
	}//<---End loop over clusters iclu
	
	
//-------------------------------------------------------------------------------------------------------------------------------------------------    
    
    
    
    
    // ##########################################
    // ### Variables to save for each cluster ###
    // ##########################################
    int   nClustersFound = 0;				//<---Number of clusters to be evaluated in a single plane 
    							//    (this gets zeroed after looping over all clusters in a plane)
							
    int   n2dVertexCandidates = 0;			//<---Number of candidate 2d Vertices found
    float Clu_Plane[10000] = {0};         	     	//<---Plane of the current cluster
    float Clu_StartPos_Wire[10000]= {0};     		//<---Starting wire number of cluster
    float Clu_StartPosUncer_Wire[10000]= {0};		//<---Starting wire number uncertainty
    float Clu_StartPos_TimeTick[10000]= {0};      	//<---Starting TDC value of the cluster
    float Clu_StartPosUncer_TimeTick[10000]= {0}; 	//<---Starting TDC value uncertainty
  
    float Clu_EndPos_Wire[10000]= {0};       		//<---Ending wire number of cluster
    float Clu_EndPosUncer_Wire[10000]= {0};  		//<---Ending wire number uncertainty
    float Clu_EndPos_TimeTick[10000]= {0};        	//<---Ending TDC value of the cluster
    float Clu_EndPosUncer_TimeTick[10000]= {0};   	//<---Ending TDC value uncertainty
  
    float Clu_Slope[10000]= {0};	  	   		//<---Calculated Slope of the cluster (TDC/Wire)
    float Clu_SlopeUncer[10000]= {0};        		//<---Slope Error
    float Clu_Yintercept[10000]= {0};			//<---Clusters Y Intercept using start positions
    float Clu_Yintercept2[10000]= {0};			//<---Clusters Y Intercept using end positions
    float Clu_Length[10000]= {0};			//<---Calculated Length of the cluster
    
    int   AllCluster = 0;
    float AllCluster_Plane[10000] = {0.};		//<---Storing the plane # for all clusters
    float AllCluster_Length[10000] = {0.};		//<---Storing the length for all clusters
    float AllCluster_StartWire[10000] = {0.};		//<---Storing the Start Wire for all clusters
    float AllCluster_StartTime[10000] = {0.};		//<---Storing the Start Time for all clusters
    float AllCluster_EndWire[10000] = {0.};		//<---Storing the End Wire for all clusters
    float AllCluster_EndTime[10000] = {0.};		//<---Storing the End Time for all clusters
    
    
    // ##############################
    // ### Looping over cryostats ###
    // ##############################
    for(size_t cstat = 0; cstat < geom->Ncryostats(); ++cstat)
    	{
    	// ##########################
      	// ### Looping over TPC's ###
      	// ##########################
      	for(size_t tpc = 0; tpc < geom->Cryostat(cstat).NTPC(); ++tpc)
		{
		nplanes = geom->Cryostat(cstat).TPC(tpc).Nplanes();
		
		// === Defining some variables to be used later JA: Update this ===
		std::vector<int> cluvtx[nplanes];
    		
    		// #################################
		// ### Loop over the wire planes ###
		// #################################
		for (int i = 0; i < geom->Cryostat(cstat).TPC(tpc).Nplanes(); ++i)
			{
    			//           geom->WirePitch(Wire1, Wire2, Plane#, TPC#, Cyro#);
    			wire_pitch = geom->WirePitch(0,1,i,tpc,cstat);
    			
			// ##############################################
	  		// ### If there is at least one cluster found ###
	  		// ##############################################
	  		if (Cls[i].size() >= 1)
				{
				// ##############################
	    			// ### Loop over each cluster ###
	    			// ##############################
	    			for (unsigned j = 0; j<Cls[i].size(); ++j)
					{
					
					// === Current Clusters Plane ===
					Clu_Plane[nClustersFound]  		  = clusters[Cls[i][j]]->View();
					AllCluster_Plane[AllCluster]		  = Clu_Plane[nClustersFound];
					
					// === Current Clusters StartPos ===
					Clu_StartPos_Wire[nClustersFound]	  = clusters[Cls[i][j]]->StartPos()[0];
					AllCluster_StartWire[AllCluster]	  = Clu_StartPos_Wire[nClustersFound];
					Clu_StartPosUncer_Wire[nClustersFound]	  = clusters[Cls[i][j]]->SigmaStartPos()[0];
					Clu_StartPos_TimeTick[nClustersFound]	  = clusters[Cls[i][j]]->StartPos()[1];
					AllCluster_StartTime[AllCluster]	  = Clu_StartPos_TimeTick[nClustersFound];
					Clu_StartPosUncer_TimeTick[nClustersFound]= clusters[Cls[i][j]]->SigmaStartPos()[1];
					
					// === Current Clusters EndPos ===
					Clu_EndPos_Wire[nClustersFound]		  = clusters[Cls[i][j]]->EndPos()[0];
					AllCluster_EndWire[AllCluster]		  = Clu_EndPos_Wire[nClustersFound];
					Clu_EndPosUncer_Wire[nClustersFound]	  = clusters[Cls[i][j]]->SigmaEndPos()[0];
					Clu_EndPos_TimeTick[nClustersFound]	  = clusters[Cls[i][j]]->EndPos()[1];
					AllCluster_EndTime[AllCluster]		  = Clu_EndPos_TimeTick[nClustersFound];
					Clu_EndPosUncer_TimeTick[nClustersFound]  = clusters[Cls[i][j]]->SigmaEndPos()[1];
					
					// === Current Clusters Slope (In Wire and Time Tick)
					Clu_Slope[nClustersFound] 		  = dtdwstart[Cls[i][j]];
					Clu_SlopeUncer[nClustersFound]		  = dtdwstartError[Cls[i][j]];
					
					
					Clu_Length[nClustersFound] 		  = std::sqrt(pow((clusters[Cls[i][j]]->StartPos()[0]-clusters[Cls[i][j]]->EndPos()[0])*13.5,2) //<---JA: Note this 13.5 is a hard coded
					 						    +pow(clusters[Cls[i][j]]->StartPos()[1]-clusters[Cls[i][j]]->EndPos()[1],2));	// number that came from ArgoNeuT MC...fix me!
					
					AllCluster_Length[AllCluster]		  = Clu_Length[nClustersFound];
					// ######################################################
					// ### Given a slope and a point find the y-intercept ###
					// ###                   c = y-mx                     ###
					// ######################################################
					Clu_Yintercept[nClustersFound] = Clu_StartPos_TimeTick[nClustersFound] - (Clu_Slope[nClustersFound] * Clu_StartPos_Wire[nClustersFound]);
					
					// #################################################################
					// ###     Also calculating the y-intercept but using the        ###
					// ###   end time of the cluster correct for the possibility     ###
					// ### that the clustering didn't get start and end points right ###
					// #################################################################
					Clu_Yintercept2[nClustersFound] = Clu_EndPos_TimeTick[nClustersFound] - (Clu_Slope[nClustersFound] * Clu_EndPos_Wire[nClustersFound]);
					
					
					nClustersFound++;
					AllCluster++;
    					}//<---End looping over clusters
					
					
    				}//<---End checking that we have at least one cluster found
			// ################################################################
			// ## If no clusters were found then put in dummy vertex values ###
			// ################################################################
			else
				{
	    			vtx_wire.push_back(-1);
	    			vtx_time.push_back(-1);
				vtx_plane.push_back(-1);
	  			}//<---End no clusters found else statement
			
			
			
			
			// ################################################################################
			// ### Now we try to find a 2-d vertex in the plane we are currently looking in ###
			// ################################################################################
			for (unsigned int n = nClustersFound; n > 0; n--)
				{
				// #######################################################
				// ###   Looping over the clusters starting from the   ###
				// ### first cluster and checking against the nCluster ###
				// #######################################################
				for (unsigned int m = 0; m < n; m++)
					{
					// ###########################################################
					// ### Checking to make sure clusters are in the same view ###
					// ###########################################################
					if(Clu_Plane[n] == Clu_Plane[m])
						{
						// --- Skip the vertex if the lines slope don't intercept ---
						if(Clu_Slope[m] - Clu_Slope[n] == 0){break;}
						
						/*
						// ============================================================
						// === X intersection = (yInt2 - yInt1) / (slope1 - slope2) ===
						float intersection_X = (Clu_Yintercept[n] - Clu_Yintercept[m]) / (Clu_Slope[m] - Clu_Slope[n]);
						// ================================================
						// === Y intersection = (slope1 * XInt) + yInt1 ===
						float intersection_Y = (Clu_Slope[m] * intersection_X) + Clu_Yintercept[m];
						
						
						// ##########################################################
						// ### Filling the vector of Vertex Wire, Time, and Plane ###
						// ##########################################################
						
						// -----------------------------------------------------------------------------
						// --- Skip this vertex if the X  and Y intersection is outside the detector ---
						// --- using geom->Nwires(plane,tpc,cyrostat) & detprop->NumberTimeSamples() ---
						// -----------------------------------------------------------------------------
						if( intersection_X > 1 && intersection_Y > 0 && 
						    (intersection_X < geom->Nwires(Clu_Plane[n],0,0) || intersection_X < geom->Nwires(Clu_Plane[m],tpc,cstat) ) &&
						    intersection_Y < detprop->NumberTimeSamples() )
							{
							
							
							vtx_wire.push_back(intersection_X);
							vtx_time.push_back(intersection_Y);
							vtx_plane.push_back(Clu_Plane[m]);
							n2dVertexCandidates++;
							}//<---End saving a "good 2d vertex" candidate
						*/	
						
						
						// #####################################################
						// ### Skipping calculating for end points to reduce ###
						// ###    the number of candidate verticies found    ###
						// #####################################################
						
						// ============================================================
						// === X intersection = (yInt2 - yInt1) / (slope1 - slope2) ===
						float intersection_X2 = (Clu_Yintercept2[n] - Clu_Yintercept2[m]) / (Clu_Slope[m] - Clu_Slope[n]);
						// ================================================
						// === Y intersection = (slope1 * XInt) + yInt1 ===
						float intersection_Y2 = (Clu_Slope[m] * intersection_X2) + Clu_Yintercept2[m];
						
						
						// #########################################
						// ### Skipping crap intersection points ###
						// #########################################
						if(intersection_X2 < 1){continue;}
						if(intersection_X2 > geom->Nwires(Clu_Plane[m],tpc,cstat)){continue;}
						if(intersection_Y2 < 0){continue;}
						if(intersection_Y2 > detprop->NumberTimeSamples() ){continue;}
						
						
						// ### If the intersection point is 80 or more wires away from either cluster
						// ### and the one of the clusters has fewer than 10 hits the intersection
						// ### is likely a crap one and we won't save this point 
						
						// ### Gathering the hits associated with the current cluster ###
						std::vector< art::Ptr<recob::Hit> > hitClu1 = fmh.at(m);
						std::vector< art::Ptr<recob::Hit> > hitClu2 = fmh.at(n);
						
						if( abs( Clu_EndPos_Wire[m] - intersection_X2 ) > 80 && hitClu1.size() < 8 ||
						    abs( Clu_EndPos_Wire[n] - intersection_X2 ) > 80 && hitClu2.size() < 8 )
						    {std::cout<<"Skipping this intersection"<<std::endl;
						    continue;}
						
						// ### If the intersection point is 50 or more wires away from either cluster
						// ### and the one of the clusters has fewer than 3 hits the intersection
						// ### is likely a crap one and we won't save this point 
						if( abs( Clu_EndPos_Wire[m] - intersection_X2 ) > 50 && hitClu1.size() < 4 ||
						    abs( Clu_EndPos_Wire[n] - intersection_X2 ) > 50 && hitClu2.size() < 4 )
						    {std::cout<<"Skipping this intersection"<<std::endl;
						    continue;}    
						    
						
						// ##################################################
						// ### Now checking if this intersection point is ###
						// ###       near a 3 (???) or more hits          ###
						// ##################################################
						
						int nOverlapHits = 0;
						// =========================
						// === Looping over hits ===
						for(size_t nHits = 0; nHits < hitListHandle->size(); nHits++)
							{
							
							// === Finding Wire/Plane associated with the hit ===
       							art::Ptr<recob::Hit> hit(hitListHandle, nHits);
       							
       							//w = hit->WireID().Wire; 
							
							// ===============================================================
							// === Skipping hits that aren't in the plane with the cluster ===
							// ===============================================================
							if(Clu_Plane[m] != hit->WireID().Plane)	{continue;}
							
							double hit_wire = hit->WireID().Wire;
							double hit_time = hit->PeakTime();
							
							if( fabs(intersection_X2 - hit_wire) < 4 && fabs(intersection_Y2 - hit_time) < 20)
								{
								nOverlapHits++;
								/*std::cout<<std::endl;
								std::cout<<"intersection_X2 = "<<intersection_X2<<std::endl;
								std::cout<<"hit_wire = "<<hit_wire<<std::endl;
								std::cout<<"intersection_Y2 = "<<intersection_Y2<<std::endl;
								std::cout<<"hit_time = "<<hit_time<<std::endl;
								std::cout<<std::endl;*/
								
								
								}//<--End checking the hit overlap with the intersection
							
							
							}//<---End nHits loop
						
						
						
						if(nOverlapHits <1)
							{continue;}
							
						//std::cout<<std::endl;
						//std::cout<<"Wire = "<<intersection_X2<<std::endl;
						//std::cout<<"Time = "<<intersection_Y2<<std::endl;
						//std::cout<<"nOverlapHits = "<<nOverlapHits<<std::endl;
						//std::cout<<std::endl;
						
						// ##########################################################
						// ### Filling the vector of Vertex Wire, Time, and Plane ###
						// ##########################################################
						
						// -----------------------------------------------------------------------------
						// --- Skip this vertex if the X  and Y intersection is outside the detector ---
						// --- using geom->Nwires(plane,tpc,cyrostat) & detprop->NumberTimeSamples() ---
						// -----------------------------------------------------------------------------
						if( intersection_X2 > 1 && intersection_Y2 > 0 && 
						    ( intersection_X2 < geom->Nwires(Clu_Plane[m],tpc,cstat) ) &&
						    intersection_Y2 < detprop->NumberTimeSamples() )
							{
							
							
							vtx_wire.push_back(intersection_X2);
							vtx_time.push_back(intersection_Y2);
							vtx_plane.push_back(Clu_Plane[m]);
							n2dVertexCandidates++;
							}//<---End saving a "good 2d vertex" candidate	
						
						
						}//<---End making sure we are in the same plane
					}//<---End m ++ loop
				}//<--- End n-- loop
			
			// === Zeroing the clusters for the current view
			nClustersFound = 0;
    			}//<---End loop over wireplanes
    		}//<---End looping over TPC's
    	}//<---End looping over cryostats




    // ########################################################################
    // ### Introducing a merge step for the candidate 2-d vertex candidates ###
    // ########################################################################
    double Wire[100000]  = {0.};
    double Time[100000]  = {0.};
    double Plane[100000] = {0.};
    
    for(unsigned int vtxloop = 0 ; vtxloop < n2dVertexCandidates; vtxloop++)
	{
	Wire[vtxloop]  = vtx_wire[vtxloop];
	Time[vtxloop]  = vtx_time[vtxloop];
	Plane[vtxloop] = vtx_plane[vtxloop];
	
	}
    
    double vtx_wire_merged[100000]  = {0.};
    double vtx_time_merged[100000]  = {0.};
    double vtx_plane_merged[100000] = {0.};
    
    double temp_wire1 = 0, temp_time1 = 0, temp_plane1 = 0;
    double temp_wire2 = 0, temp_time2 = 0, temp_plane2 = 0;
    bool merged = false;
    
    int n2dMergedVertices = 0;
    

    // #########################################
    // ### Looping over 2d-verticies (loop1) ###
    // #########################################
    for(unsigned int vtxloop1 = 0 ; vtxloop1 < n2dVertexCandidates; vtxloop1++)
	{
	if(Wire[vtxloop1] < 0){continue;}
	
	//std::cout<<"Wire[vtxloop1] = "<<Wire[vtxloop1]<<std::endl;
	
	temp_wire1  = Wire[vtxloop1];
	temp_time1  = Time[vtxloop1];
	temp_plane1 = Plane[vtxloop1];
	
	// #########################################
	// ### Looping over 2d-verticies (loop2) ###
	// #########################################
	for(unsigned int vtxloop2 = n2dVertexCandidates; vtxloop2 > vtxloop1; vtxloop2--)
    		{
		if(Wire[vtxloop2] < 0){continue;} 
		
		
		temp_wire2  = Wire[vtxloop2];
		temp_time2  = Time[vtxloop2];
		temp_plane2 = Plane[vtxloop2];
		
		merged = false;
		// ########################################################
		// ### Make sure the 2d-Verticies are in the same plane ###
		// ########################################################
		if(temp_plane1 == temp_plane2)
			{
			// ###############################################			
			// ### Considering merging 2d vertices if they ###
			// ###    are within 3 wires of each other     ###
			// ###############################################
			if( fabs(temp_wire1 - temp_wire2) < 4 )
				{
				// ############################################################
				// ### Merge the verticies if they are within 15 time ticks ###
				// ############################################################
				if( fabs(temp_time1 - temp_time2) < 16 )
					{
					vtx_wire_merged[n2dMergedVertices] = ((temp_wire2 + temp_wire1)/ 2) ;
					vtx_time_merged[n2dMergedVertices] = ((temp_time2 + temp_time1)/ 2) ;
					vtx_plane_merged[n2dMergedVertices] = temp_plane2;
					
					
					merged = true;
					Wire[vtxloop1] = -999;
					Time[vtxloop1] = -999;
					Plane[vtxloop1] = -999;
					
					Wire[vtxloop2] = -999;
					Time[vtxloop2] = -999;
					Plane[vtxloop2] = -999;
					}//<---End the check within 15 time ticks for merging
				}//<---Looking at vertices that are within 3 wires of each other
			}//<---Only looking at vertices that are in the same plane
		
		}//<---End vtxloop2
	if(!merged)
		{
		vtx_wire_merged[n2dMergedVertices]  = temp_wire1;
		vtx_time_merged[n2dMergedVertices]  = temp_time1 ;
		vtx_plane_merged[n2dMergedVertices] = temp_plane1;

		}//<---end saving unmerged verticies
			
	//std::cout<<"vtx_wire_merged[n2dMergedVertices] = "<<vtx_wire_merged[n2dMergedVertices]<<std::endl;
	n2dMergedVertices++;
	}//<---End vtxloop1
	
    
    
    //std::cout<<std::endl;
    //std::cout<<"n2dVertexCandidates = "<<n2dVertexCandidates<<std::endl;
    //std::cout<<"n2dMergedVertices   = "<<n2dMergedVertices<<std::endl;
    //std::cout<<std::endl;

    
    double y_coord = 0, z_coord = 0;
    
    double x_3dVertex[100000] = {0.}, y_3dVertex[100000] = {0.}, z_3dVertex[100000] = {0.};
    int n3dVertex = 0;
    uint32_t     vtx1_channel = 0;
    uint32_t     vtx2_channel = 0;
    
    // --------------------------------------------------------------------------
    // ---   Having now found a very long list of potential 2-d end points    ---
    // --- we need to check if any of them match between planes and only keep ---
    // ---                       those that have matches                      ---
    // --------------------------------------------------------------------------
    
    // ##############################
    // ### Looping over cryostats ###
    // ##############################
    for(size_t cstat = 0; cstat < geom->Ncryostats(); ++cstat)
    	{
    	// ##########################
      	// ### Looping over TPC's ###
      	// ##########################
      	for(size_t tpc = 0; tpc < geom->Cryostat(cstat).NTPC(); ++tpc)
		{
    		for(unsigned int vtx = n2dMergedVertices; vtx > 0; vtx--)
    			{
			for (unsigned int vtx1 = 0; vtx1 < vtx; vtx1++)
				{
				// ###########################################################################
				// ### Check to make sure we are comparing verticies from different planes ###
				// ###########################################################################
				if(vtx_plane_merged[vtx1] != vtx_plane_merged[vtx])
					{
					/*std::cout<<std::endl;
					std::cout<<"vtx_wire[vtx1] = "<<vtx_wire[vtx1]<<" , vtx_wire[vtx] = "<<vtx_wire[vtx]<<std::endl;
					std::cout<<"vtx_time[vtx1] = "<<vtx_time[vtx1]<<" , vtx_time[vtx] = "<<vtx_time[vtx]<<std::endl;
					std::cout<<"vtx_plane[vtx1] = "<<vtx_plane[vtx1]<<" , vtx_plane[vtx] = "<<vtx_plane[vtx]<<std::endl;
					std::cout<<std::endl;*/
			
					// === To figure out if these two verticies are from a common point 
					// === we need to check if the channels intersect and if they are 
					// === close in time ticks as well...to do this we have to do some 
					// === converting to use geom->PlaneWireToChannel(PlaneNo, Wire, tpc, cstat)
					// === JA: Need to include vtx tpc, and cstat to make detector agnositc
			
					unsigned int vtx1_plane   = vtx_plane_merged[vtx1];
					unsigned int vtx1_wire    = vtx_wire_merged[vtx1];
					try
						{
						vtx1_channel = geom->PlaneWireToChannel(vtx1_plane, vtx1_wire, tpc, cstat);
				
						}
					catch(...)
						{
						mf::LogWarning("FeatureVertexFinder") << "PlaneWireToChannel Failed";
	      
	      					continue;
	    					}
				
				
			
					unsigned int vtx2_plane   = vtx_plane_merged[vtx];
					unsigned int vtx2_wire    = vtx_wire_merged[vtx];
			
					try
						{
						vtx2_channel = geom->PlaneWireToChannel(vtx2_plane, vtx2_wire, tpc, cstat);
				
						}
					catch(...)
						{
						mf::LogWarning("FeatureVertexFinder") << "PlaneWireToChannel Failed";
	      
	      					continue;
				
						}
			
					// ##############################################################################
					// ### Check to see if the channels intersect and save the y and z coordinate ###
					// ##############################################################################
					bool match = false;
			
					try
						{
						match = geom->ChannelsIntersect( vtx1_channel, vtx2_channel, y_coord, z_coord);
				
				
						}
					catch(...)
						{
	      					mf::LogWarning("FeatureVertexFinder") << "match failed for some reason";
	      
	      					match = false;
	      					continue;
	    					}
					if( match )
						{
						//std::cout<<std::endl;
						//std::cout<<"Vertex Intersection Found!!!"<<std::endl;
						//std::cout<<"Offset between planes = "<<detprop->TimeOffsetV()<<std::endl;
						//std::cout<<std::endl;
				
						// #############################################################################
						// ### Now check if the matched channels are within 1 cm when projected in X ###
						// #############################################################################
						float tempXCluster1 = detprop->ConvertTicksToX(vtx_time_merged[vtx1], vtx1_plane, tpc, cstat);
						float tempXCluster2 = detprop->ConvertTicksToX(vtx_time_merged[vtx], vtx2_plane, tpc, cstat);
						
						if(std::abs(tempXCluster1 - tempXCluster2) < 0.5)
							{
							//        detprop->ConvertTicksToX(ticks, plane, tpc, cryostat)
							x_3dVertex[n3dVertex] = detprop->ConvertTicksToX(vtx_time_merged[vtx1], vtx_plane_merged[vtx1], tpc, cstat); 
							y_3dVertex[n3dVertex] = y_coord;
							z_3dVertex[n3dVertex] = z_coord;
							n3dVertex++;
					
					
							//std::cout<<std::endl;
							//std::cout<<" ============== REAL 3-D VERTEX FOUND ============"<<std::endl;
							//std::cout<<" std::abs(vtx_time[vtx1] - vtx_time[vtx] = "<<std::abs(vtx_time[vtx1] - vtx_time[vtx])<<std::endl;
							//std::cout<<"x_coord = "<<x_coord<<" , y_coord = "<<y_coord<<" , z_coord = "<<z_coord<<std::endl;
							//std::cout<<std::endl;
					
					
					
					
							}//<---End Checking if the vertices agree "well enough" in time tick
				
						}//<---End Checking if verticies intersect

					}//<--- End checking we are in different planes
				}//<---end vtx1 for loop
			}//<---End vtx for loop
		}//<---End loop over TPC's 
	}//<---End loop over cryostats








//-----------------------------------------------------------------------------------------------------------------------------

// ###################################################################
// ### Now we need to make sure that we remove duplicate verticies ###
// ###    just in case we have any in our list of n3dVertex        ###
// ###################################################################

double x_3dVertex_dupRemoved[100000] = {0.}, y_3dVertex_dupRemoved[100000] = {0.}, z_3dVertex_dupRemoved[100000] = {0.};
double strength_3dVertex_dubRemoved[10000] = {0.};
int n3dVertex_dupRemoved = 0;

for(size_t dup = 0; dup < n3dVertex; dup ++)
	{
	float tempX_dup = x_3dVertex[dup];
	float tempY_dup = y_3dVertex[dup];
	float tempZ_dup = z_3dVertex[dup];
	
	
	bool duplicate_found = false;
	
	for(size_t check = dup+1; check < n3dVertex; check++)
		{
		
		// #############################################################################
		// ### I am going to call a duplicate vertex one that matches in x, y, and z ###
		// ###        within 0.1 cm for all 3 coordinates simultaneously             ###
		// #############################################################################
		if(std::abs( x_3dVertex[check] - tempX_dup ) < 0.1 && std::abs( y_3dVertex[check] - tempY_dup ) < 0.1 &&
		   std::abs( z_3dVertex[check] - tempZ_dup ) < 0.1 )
		   	{
			duplicate_found = true;
			
			}//<---End checking to see if this is a duplicate vertex
	
	
	
		}//<---End check for loop
	
	// ######################################################################
	// ### If we didn't find a duplicate then lets save this 3d vertex as ###
	// ###            a real candidate for consideration                  ###
	// ######################################################################
	if(!duplicate_found)
		{
		x_3dVertex_dupRemoved[n3dVertex_dupRemoved] = tempX_dup;
		y_3dVertex_dupRemoved[n3dVertex_dupRemoved] = tempY_dup;
		z_3dVertex_dupRemoved[n3dVertex_dupRemoved] = tempZ_dup;
		
		n3dVertex_dupRemoved++;
		}	

	}//<---End dup for loop




/*std::cout<<std::endl;
std::cout<<"#####################################"<<std::endl;
std::cout<<std::endl;
std::cout<<" N3dVertex w/ duplicates removed = "<<n3dVertex_dupRemoved<<std::endl;
std::cout<<std::endl;
std::cout<<"#####################################"<<std::endl;*/
std::cout<<std::endl;
//-----------------------------------------------------------------------------------------------------------------------------

       	// ######################################################################################################
       	// ##################### Now I am going to deal with the 3 cases for finding 3d verticies ###############
       	// ### Case1) If n3dVertex_dupRemoved == 1: Then I only found one intersection point using the clusters found,
       	// ###                           in this case I will simply save to the event a EndPoint2d and a 3d
       	// ###                           vertex (JA: Need to add relevant associations later)
	// ###
       	// ### Case2) If n3dVertex_dupRemoved > 1: In this case I will attempt to find a 3d feature that is within 2cm
       	// ###                         		   in x, y, and z. If there is more than one of these points that match
       	// ###                          	   I will, for now (JA: Need to improve), take the one with the smaller 
       	// ###                          	   z coordinate (more upstream)
       	// ###
       	// ### Case3) If n3dVertex_dupRemoved = 0: Still trying to figure out what to do here....JA:(will come back to)
       	// #######################################################################################################
       

       
double TwoDvertexStrength = 0;
       	// ### Case 1...only one 3d vertex found ###
       	if (n3dVertex_dupRemoved == 1)
		{
		//std::cout<<" ### In case 1 ###"<<std::endl;
		
		// ##############################
    		// ### Looping over cryostats ###
    		// ##############################
    		for(size_t cstat = 0; cstat < geom->Ncryostats(); ++cstat)
    			{
    			// ##########################
      			// ### Looping over TPC's ###
      			// ##########################
      			for(size_t tpc = 0; tpc < geom->Cryostat(cstat).NTPC(); ++tpc)
				{
    				// #################################
				// ### Loop over the wire planes ###
				// #################################
				for (size_t i = 0; i < geom->Cryostat(cstat).TPC(tpc).Nplanes(); ++i)
					{
					double xyz[3] = {x_3dVertex_dupRemoved[0], y_3dVertex_dupRemoved[0], z_3dVertex_dupRemoved[0]};
					
					// ########################################################
					// ### Give the current 3d vertex found a strength of 1 ###
					// ### We will add +1 to it for each 3d feature that is ###
					// ###              near by it in x, y, z space         ###
					// ########################################################
					TwoDvertexStrength = 1;
					
					// #########################################################
					// ### Does this point correspond to a 3d-Feature Point? ###
					// ###          Loop over all the feature points         ###
					// #########################################################
					for (int a = 0 ; a < n3dFeatures; a++)
						{
						// #############################################################
						// ### If you find a feature point within 3 cm of the vertex ###
						// ###           add a point to the vertex strength          ###
						// #############################################################
						if (  std::abs(xyz[0] - x_feature[a]) < 3 && 
						      std::abs(xyz[1] - y_feature[a]) < 3 && 
						      std::abs(xyz[2] - z_feature[a]) < 3 )
							{
							TwoDvertexStrength++;
							//std::cout<<"Adding strength to the vertex in case 1"<<std::endl;
							//std::cout<<std::endl;
						
						
							}
						}//<---End a for loop
					
					double EndPoint2d_TimeTick = detprop->ConvertXToTicks(xyz[0],i, tpc, cstat);
					int EndPoint2d_Wire = 0;
					int EndPoint2d_Channel     = 0;
					
					try
					
						{EndPoint2d_Wire 	   = geom->NearestWire(xyz , i, tpc, cstat);}
					catch(...)
						{
						mf::LogWarning("FeatureVertexFinder") << "2dWire failed";
	    
	      					continue;
						
						}
					try
						{EndPoint2d_Channel     = geom->NearestChannel(xyz, i, tpc, cstat);}
						
					catch(...)
						{
						mf::LogWarning("FeatureVertexFinder") << "2dWire failed";
	    
	      					continue;
						
						}
						
					geo::View_t View	   = geom->View(EndPoint2d_Channel);
					geo::WireID wireID(cstat,tpc,i,EndPoint2d_Wire);
					
					
					// ### Saving the 2d Vertex found ###
					recob::EndPoint2D vertex( EndPoint2d_TimeTick , //<---TimeTick
								  wireID ,		//<---geo::WireID
								  TwoDvertexStrength ,	//<---Vtx strength (JA: ?)
								  epcol->size() ,	//<---Vtx ID (JA: ?)
								  View ,		//<---Vtx View 	
								  1 );			//<---Total Charge (JA: Need to figure this one?)
					epcol->push_back(vertex);	
					
					
							   
					}//<---End loop over Planes
				}//<---End loop over tpc's
			}//<---End loop over cryostats
			
		// ############################
		// ### Saving the 3d vertex ###
		// ############################
		double xyz2[3] = {x_3dVertex_dupRemoved[0], y_3dVertex_dupRemoved[0], z_3dVertex_dupRemoved[0]};
		recob::Vertex the3Dvertex(xyz2, vcol->size());
		vcol->push_back(the3Dvertex);
		
		
		}//<---End Case 1, only one 3d Vertex found

//----------------------------------------------------------------------------------------------------------------------------- 
	// ###      Case 2...multiple 3d verticies found       ###
	// ### To break the degeneracy start by looping over   ###
	// ###    the 3d verticies found and merge any verticies #
	// ### that are near to each other, then loop over th  ###
	// ### 3d feature points found adding strength to the  ###
	// ###     vertex that are within 3 cm of the vertex   ###
	int totalGood = 0;
	int nMerges = 0;
	double x_good[10000] = {0.}, y_good[10000] = {0.}, z_good[10000] = {0.}, strength_good[10000] = {0.};
	
	int totalGood2 = 0;
	double x_good2[10000] = {0.}, y_good2[10000] = {0.}, z_good2[10000] = {0.}, strength_good2[10000] = {0.};
	
	int match2planes = 0;
	
	//std::cout<<" n3dVertex = "<<n3dVertex<<std::endl;
	if (n3dVertex_dupRemoved > 1)
		{
		//std::cout<<" ### In case 2 ###"<<std::endl;
		// ######################################################
		// ###     Setting a limit to the number of merges    ###
		// ### to be 3 times the number of 3d verticies found ###
		// ######################################################
		int LimitMerge = 3 * n3dVertex_dupRemoved;
		// ### Trying to merge nearby verticies found ###
		for( int merge1 = 0; merge1 < n3dVertex_dupRemoved; merge1++)
			{
			TwoDvertexStrength = 1;  
			for( int merge2 = n3dVertex ; merge2 > merge1; merge2--)
				{
				double temp1_x = x_3dVertex[merge1];
				double temp1_y = y_3dVertex[merge1];
				double temp1_z = z_3dVertex[merge1];
				
				double temp2_x = x_3dVertex[merge2];
				double temp2_y = y_3dVertex[merge2];
				double temp2_z = z_3dVertex[merge2];
				
				if( temp1_x == 0 || temp1_y == 0|| temp1_z == 0 ||
				    temp2_x == 0 || temp2_y == 0 || temp2_z == 0) {continue;}
				    
				/*std::cout<<std::endl;
				std::cout<<" Should I merge these? "<<std::endl;
				std::cout<<"temp1_x = "<<temp1_x<<" , temp1_y = "<<temp1_y<<" , temp1_z = "<<temp1_z<<std::endl;
				std::cout<<"temp2_x = "<<temp2_x<<" , temp2_y = "<<temp2_y<<" , temp2_z = "<<temp2_z<<std::endl;
				std::cout<<std::endl;*/
				
				// ### Merge the verticies if they are within 1.0 cm of each other ###
				if ( (std::abs( temp1_x - temp2_x ) < 2.0 && temp1_x != 0 && temp2_x !=0) &&
				     (std::abs( temp1_y - temp2_y ) < 2.0 && temp1_y != 0 && temp2_y !=0) &&
				     (std::abs( temp1_z - temp2_z ) < 2.0 && temp1_z != 0 && temp2_z !=0) &&
				      nMerges < LimitMerge)
				    	{
					//std::cout<<" Yup, I am going to merge these! "<<std::endl;
					//std::cout<<" I've now performed a merge "<<nMerges<<" times"<<std::endl;
					//std::cout<<"Merging"<<std::endl;
					//std::cout<<"n3dVertex = "<<n3dVertex<<std::endl;
					nMerges++;
					// ### Zero the vertex that I am merging ###
					x_3dVertex[merge2] = 0.0;
					y_3dVertex[merge2] = 0.0;
					z_3dVertex[merge2] = 0.0;
					
					n3dVertex_dupRemoved++;
					
					//std::cout<<"n3dVertex now equals = "<<n3dVertex<<std::endl;
					// ### Add the merged vertex to the end of the vector ###
					x_3dVertex[n3dVertex_dupRemoved] = (temp1_x + temp2_x)/2;
					y_3dVertex[n3dVertex_dupRemoved] = (temp1_y + temp2_y)/2;
					z_3dVertex[n3dVertex_dupRemoved] = (temp1_z + temp2_z)/2;
					
					// ######################################################################
					// ### If we merged the verticies then increase its relative strength ###
					// ######################################################################
					TwoDvertexStrength++;
					strength_3dVertex_dubRemoved[n3dVertex_dupRemoved] = TwoDvertexStrength;
					//std::cout<<" The new merged vertex = "<<x_3dVertex[n3dVertex]<<" , "<<y_3dVertex[n3dVertex]<<" , "<<z_3dVertex[n3dVertex]<<std::endl;
					
					
					}//<---End merging verticies
				}//<---End merge2 loop
			}//<---End merge1 loop
			
		// ######################################################################
		// ### Now loop over the new list and only save the non-zero vertices ###
		// ######################################################################
		for (int goodvtx = 0; goodvtx < n3dVertex; goodvtx++)
			{
			if( x_3dVertex[goodvtx] != 0.0 && 
			    y_3dVertex[goodvtx] != 0.0 &&
			    z_3dVertex[goodvtx] != 0.0 )
			    	{
				bool duplicate = false;
				// ###############################################
				// ### Check to make sure this isn't a copy of ###
				// ###        a previously found vertex        ###
				// ###############################################
				for (int check = goodvtx+1; check > 0; check--)
					{
					// #########################################################################
					// ### Adding a protection that we don't check the vertex against itself ###
					// #########################################################################
					if (check != goodvtx)
						{
						// ### check if this vertex exists in the list ###
						if (x_3dVertex[goodvtx] == x_3dVertex[check] &&
						    y_3dVertex[goodvtx] == y_3dVertex[check] &&
						    z_3dVertex[goodvtx] == z_3dVertex[check] )
						    {
						    duplicate = true;
					    
						    }//<---End duplicate 
						}//<---Don't check the vertex against itself
					
					}//<---end check loop
				if(!duplicate)
					{	
					x_good[totalGood] =  x_3dVertex[goodvtx];
					y_good[totalGood] =  y_3dVertex[goodvtx];
					z_good[totalGood] =  z_3dVertex[goodvtx];
					strength_good[totalGood] = strength_3dVertex_dubRemoved[goodvtx];
					totalGood++;
					}//<---End removing duplicates
			
				}
			}//<---End goodvtx loop
		//std::cout<<"totalGood = "<<totalGood<<std::endl;
		
		
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// ###################################################
		// ### Loop over good verticies to only save those ###
		// ###       that match some number of hits        ###
		// ###################################################
		
		// ##############################
		// ### Looping over verticies ###
		// ##############################
		for(int ll = 0; ll < totalGood; ll++)
			{
			
			// ##############################
    			// ### Looping over cryostats ###
    			// ##############################
    			for(size_t cstat = 0; cstat < geom->Ncryostats(); ++cstat)
    				{
    				// ##########################
      				// ### Looping over TPC's ###
      				// ##########################
      				for(size_t tpc = 0; tpc < geom->Cryostat(cstat).NTPC(); ++tpc)
					{
    					// #################################
					// ### Loop over the wire planes ###
					// #################################
					for (size_t i = 0; i < geom->Cryostat(cstat).TPC(tpc).Nplanes(); ++i)
						{
						double xyz[3] = {x_good[ll], y_good[ll], z_good[ll]};
						double EndPoint2d_TimeTick = detprop->ConvertXToTicks(x_good[ll],i, tpc, cstat);
						int EndPoint2d_Wire = 0;
						try
					
							{EndPoint2d_Wire 	   = geom->NearestWire(xyz , i, tpc, cstat);}
						catch(...)
							{
							mf::LogWarning("FeatureVertexFinder") << "2dWire failed";
	    
	      						continue;
						
							}

						// #############################################################
						// ### Only going to save the vertex if it is within 3 wires ###
						// ###       and 20 time ticks of a hit in the event         ###
						// #############################################################
						
						int matchedVtxToHit = 0;
						// =========================
						// === Looping over hits ===
						for(size_t nHits = 0; nHits < hitListHandle->size(); nHits++)
							{
							
							// === Finding Wire/Plane associated with the hit ===
       							art::Ptr<recob::Hit> hit(hitListHandle, nHits);
       							
							
							// ===============================================================
							// === Skipping hits that aren't in the plane with the cluster ===
							// ===============================================================
							if(i != hit->WireID().Plane)	{continue;}
							
							double hit_wire = hit->WireID().Wire;
							double hit_time = hit->PeakTime();
							
							if( fabs(EndPoint2d_Wire - hit_wire) < 4 && fabs(EndPoint2d_TimeTick - hit_time) < 20)
								{
								matchedVtxToHit++;
								
								
								}
								
							}//<---End loop over hits
						
						// #####################################################
						// ### Saving the 2d Vertex found if matched to more ###
						// ### hits than what you would expect along a track ###
						// #####################################################
						if(matchedVtxToHit > 5)
							{
							match2planes++;
							}	

						}//<---End loop over Planes
					}//<---End loop over tpc's
				}//<---End loop over cryostats
			// #########################################
			// ### Saving the 3d vertex if there was ###
			// ###    a match in at least 2 planes   ###
			// #########################################
			if(match2planes > 1)
				{
				x_good2[totalGood2] = x_good[ll];
				y_good2[totalGood2] = y_good[ll];
				z_good2[totalGood2] = z_good[ll];
				totalGood2++;
				
				}	

			match2planes = 0;
			
			}//<---End ll loop
		///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		
		
		// ##############################
		// ### Looping over verticies ###
		// ##############################
		for(int l = 0; l < totalGood2; l++)
			{
			// #############################
			// ### Looping over features ###
			// #############################
			for (int f = 0; f < n3dFeatures; f++)
				{
				// ###########################################################
				// ### Looking for features and verticies within 3cm in 3d ###
				// ###########################################################
				if(std::abs(x_good2[l] - x_feature[f]) <= 3 &&  
				   std::abs(y_good2[l] - y_feature[f]) <= 3 &&
				   std::abs(z_good2[l] - z_feature[f]) <= 3)
				   	{
					
					// ##########################################################
					// ### If there is also a 3d feature near this point then ###
					// ###      increase the vertex strength again            ###
					// ##########################################################
					
					strength_good[l] = strength_good[l]+1;
					
					
					//std::cout<<std::endl;
					//std::cout<<" ###########################################"<<std::endl;
					//std::cout<<" ### 3d match between feature and vertex ###"<<std::endl;
					//std::cout<<std::endl;
					//std::cout<<"x_3dVertex[l] = "<<x_3dVertex[l]<<" , y_3dVertex[l] = "<<y_3dVertex[l]<<" , z_3dVertex[l] = "<<z_3dVertex[l]<<std::endl;
					//std::cout<<std::endl;
					
					}//<---End finding a match between features and verticies 
				}//<--- End feature for loop
				
			// ##############################
    			// ### Looping over cryostats ###
    			// ##############################
    			for(size_t cstat = 0; cstat < geom->Ncryostats(); ++cstat)
    				{
    				// ##########################
      				// ### Looping over TPC's ###
      				// ##########################
      				for(size_t tpc = 0; tpc < geom->Cryostat(cstat).NTPC(); ++tpc)
					{
    					// #################################
					// ### Loop over the wire planes ###
					// #################################
					for (size_t i = 0; i < geom->Cryostat(cstat).TPC(tpc).Nplanes(); ++i)
						{
						double xyz[3] = {x_good2[l], y_good2[l], z_good2[l]};
						double EndPoint2d_TimeTick = detprop->ConvertXToTicks(x_good2[l],i, tpc, cstat);
						int EndPoint2d_Wire = 0;
						int EndPoint2d_Channel = 0;
						try
					
							{EndPoint2d_Wire 	   = geom->NearestWire(xyz , i, tpc, cstat);}
						catch(...)
							{
							mf::LogWarning("FeatureVertexFinder") << "2dWire failed";
	    
	      						continue;
						
							}
							
						try
							{EndPoint2d_Channel     = geom->NearestChannel(xyz, i, tpc, cstat);}
						catch(...)
							{
							mf::LogWarning("FeatureVertexFinder") << "2dWire failed";
	    
	      						continue;
						
							}
							
						geo::View_t View	   = geom->View(EndPoint2d_Channel);
						geo::WireID wireID(cstat,tpc,i,EndPoint2d_Wire);
						
						int EndPoint2dstrength = strength_good[l];
						
						recob::EndPoint2D vertex( EndPoint2d_TimeTick , //<---TimeTick
									  wireID ,		//<---geo::WireID
									  EndPoint2dstrength ,	//<---Vtx strength (JA: ?)
									  epcol->size() ,	//<---Vtx ID (JA: ?)
									  View ,		//<---Vtx View 	
									  1 );			//<---Vtx Total Charge (JA: Need to figure this one?)
						epcol->push_back(vertex);
							

						}//<---End loop over Planes
					}//<---End loop over tpc's
				}//<---End loop over cryostats
			// #########################################
			// ### Saving the 3d vertex if there was ###
			// ###    a match in at least 2 planes   ###
			// #########################################
			
			double xyz2[3] = {x_good2[l], y_good2[l], z_good2[l]};
			recob::Vertex the3Dvertex(xyz2, vcol->size());
			vcol->push_back(the3Dvertex);
				

			}//<--End l for loop
		
		}//<---End Case 2, many 3d Verticies found
		
		
		
//----------------------------------------------------------------------------------------------------------------------------- 
	
	// ### This is for the case where we didn't find a good match
	// ### in 3d for a vertex...still need to figure out what to do here
	
	// Thing to try: Lets first try looping over the clusters and finding the longest cluster a
	// view and try matching that to cluster to a 3d-feature point...if there is a match then record
	// that as a vertex
	
	// Thing to also try: Taking the largest score 3d-feature point score and simply saving that...
	// this shifts the problem to simply finding the feature point...not terribly helpful
	
	double x_featClusMatch = 0, y_featClusMatch = 0, z_featClusMatch = 0;
	
	double LongestClusterStartWire[100] = {0};
	double LongestClusterStartTime[100] = {0};
	if (n3dVertex_dupRemoved == 0)
		{
		bool NothingFoundYet = true;
		//std::cout<<" ### We are in case 3 ###"<<std::endl;
		//std::cout<<" n3dFeatures = "<<n3dFeatures<<std::endl;
		// ##############################
    		// ### Looping over cryostats ###
    		// ##############################
    		for(size_t cstat = 0; cstat < geom->Ncryostats(); ++cstat)
    			{
    			// ##########################
      			// ### Looping over TPC's ###
      			// ##########################
      			for(size_t tpc = 0; tpc < geom->Cryostat(cstat).NTPC(); ++tpc)
				{
    				// #################################
				// ### Loop over the wire planes ###
				// #################################
				for (size_t plane = 0; plane < geom->Cryostat(cstat).TPC(tpc).Nplanes(); ++plane)
					{
					// ############################################################################
					// ### We reinitialize the strength for each plane since we haven't checked ###
					// ###  at the top of this loop if this has been matched to anything else   ###
					// ############################################################################
					TwoDvertexStrength = 0;
					
					// === Defining a bunch of temp variables ===
					float Length_LongestClusterInPlane = 0;
					float StartWire_LongestClusterInPlane = 0;
					float StartTime_LongestClusterInPlane = 0;
					float EndWire_LongestClusterInPlane = 0;
					float EndTime_LongestClusterInPlane = 0;
					float Plane_LongestClusterInPlane = 0;
					//std::cout<<" AllCluster = "<< AllCluster<<std::endl;
					// #######################################################################
					// ### Looping over clusters to find the longest cluster in each plane ###
					// #######################################################################
					for (int clus = 0; clus < AllCluster; clus++)
						{
						//std::cout<< "AllCluster_Plane[AllCluster] = "<<AllCluster_Plane[AllCluster]<<" , plane = "<<plane<<std::endl;
						// ### Checking that the current cluster is in the current plane ###
						if(AllCluster_Plane[clus] == plane)
							{
							
							// Storing the current cluster length
							float tempClusterLength = AllCluster_Length[clus];
							
							//std::cout<<std::endl;
							//std::cout<<"tempClusterLength = "<<tempClusterLength<<std::endl;
							//std::cout<<"Clu_Length[clus] = "<<AllCluster_Length[clus]<<std::endl;
							//std::cout<<"plane = "<<plane<<std::endl;
							//std::cout<<std::endl;
							
							// Is the current cluster the longest one found?
							if( tempClusterLength > Length_LongestClusterInPlane)
								{
								// Save the longest clusters information in this view
								Length_LongestClusterInPlane    = tempClusterLength;
								StartWire_LongestClusterInPlane = AllCluster_StartWire[clus];
								StartTime_LongestClusterInPlane = AllCluster_StartTime[clus];
								EndWire_LongestClusterInPlane   = AllCluster_EndWire[clus];
								EndTime_LongestClusterInPlane   = AllCluster_EndTime[clus];
								Plane_LongestClusterInPlane     = AllCluster_Plane[clus];
								}//<---End saving the longest cluster
							}//<---Current cluster in current plane	
						}///<---End clus for loop
					//std::cout<<"Longest Cluster = "<<Length_LongestClusterInPlane<< ", Plane = "<<Plane_LongestClusterInPlane<<std::endl;
					//std::cout<<"Start Wire = "<<StartWire_LongestClusterInPlane<<" , Start Time = "<<StartTime_LongestClusterInPlane<<std::endl;
					
					// #####################################################################################
					// ###       Now that we have the longest cluster in this plane lets loop over       ###
					// ### the 3d-Feature points found and see if any are consistant with this 2-d point ###
					// #####################################################################################
					
					NothingFoundYet = true;
					
					// ### Set the score for this 2d vertex in this plane to be = 0 ###
					// ### This will increase for each feature it finds that matches ###
					// ### down to it
					
					/// #### Get the 3d feature point, project it back down to the current plane...check to see if consistant
					float temp_feat_score = 0;
					
					for (int feat = 0; feat < n3dFeatures; feat++)
						{
						// #########################################################
						// ### Finding the nearest wire for this 3dfeature point ###
						// #########################################################
						float xyzfeat[3]   = {x_feature[feat],y_feature[feat],z_feature[feat]};
						float xyzfeat_score = strength_feature[feat];
						double feat_2dWire = -1;
						double feat_TimeT = -1;
						
						try
							{
							feat_2dWire = geom->NearestWire(xyzfeat, plane, tpc, cstat);
							feat_TimeT  = detprop->ConvertXToTicks(x_feature[feat], plane, tpc, cstat);
							}
						catch(...)
							{
							mf::LogWarning("FeatureVertexFinder") << "2dWire failed";
							continue;
							}
						
						//std::cout<<" Feat_2dWire = "<<feat_2dWire<<std::endl;
						//std::cout<<" Feat_timeT  = "<<feat_TimeT<<std::endl;
						// ######################################################################################
						// ###           Now check to see if this feature point is within 2 wires and         ###
						// ###   5 time ticks of the start time of the longest cluster found in this plane    ###
						// ###                and is the strongest xyz feature in the list                    ###
						// ######################################################################################
						if( std::abs(feat_2dWire - StartWire_LongestClusterInPlane) <= 2 && std::abs(StartTime_LongestClusterInPlane - feat_TimeT ) <=5 &&
						    xyzfeat_score >= temp_feat_score  )
							{
							temp_feat_score = xyzfeat_score;
							// ###########################################################
							// ### If a feature matches the 2d point increase strength ###
							// ###########################################################
							TwoDvertexStrength++;

							
							if(NothingFoundYet)
								{
								// ### Vertex found ###
								//std::cout<<std::endl;
								//std::cout<<" ##### Longest Cluster Feature Match Start Position #####"<<std::endl;
								//std::cout<<"X = "<<xyzfeat[0]<<" , Y = "<<xyzfeat[1]<<" , Z = "<<xyzfeat[2]<<std::endl;
								//std::cout<<"Plane = "<<plane<<" , Wire = "<<feat_2dWire<<" , Time = "<<feat_TimeT<<std::endl;
								//std::cout<<std::endl;
								
								x_featClusMatch = xyzfeat[0];
								y_featClusMatch = xyzfeat[1];
								z_featClusMatch = xyzfeat[2];
								
								double xyz[3] = {xyzfeat[0], xyzfeat[1], xyzfeat[2]};
								double EndPoint2d_TimeTick = detprop->ConvertXToTicks(xyzfeat[0],plane, tpc, cstat);
						
								int EndPoint2d_Wire 	   = 0;
								int EndPoint2d_Channel     = 0;
								
								try
					
									{EndPoint2d_Wire 	   = geom->NearestWire(xyz , plane, tpc, cstat);}
								catch(...)
									{
									mf::LogWarning("FeatureVertexFinder") << "2dWire failed";
	      								continue;
						
									}
							
								try
									{EndPoint2d_Channel     = geom->NearestChannel(xyz, plane, tpc, cstat);}
								catch(...)
									{
									mf::LogWarning("FeatureVertexFinder") << "2dWire failed";
	    
	      								continue;
						
									}
								geo::View_t View	   = geom->View(EndPoint2d_Channel);
								geo::WireID wireID(cstat,tpc,plane,EndPoint2d_Wire);
								
								// ### Saving the 2d Vertex found ###
								recob::EndPoint2D vertex( EndPoint2d_TimeTick , //<---TimeTick
										  wireID ,		//<---geo::WireID
										  TwoDvertexStrength ,			//<---Vtx strength (JA: ?)
										  epcol->size() ,	//<---Vtx ID (JA: ?)
										  View ,		//<---Vtx View 	
										  1 );			//<---Vtx charge (JA: Need to figure this one?)
								epcol->push_back(vertex);
								
								
								}//<---End saving this vertex found
							
							NothingFoundYet = false;
							}//<---End finding a match between the start time and the feature
						}//<---End feat for loop
					
					// ##################################################################################
					// ### After checking all the start times against the feature points, if we still ###
					// ###    don't have a vertex found check the end times of the longest cluster    ###
					// ###           (This should protect {some} against the clustering               ###
					// ###               putting the start and end in the wrong spot)                 ###
					// ##################################################################################
					if (NothingFoundYet )
						{
						for (int feat2 = 0; feat2 < n3dFeatures; feat2++)
							{
							// #########################################################
							// ### Finding the nearest wire for this 3dfeature point ###
							// #########################################################
							float xyzfeat2[3]   = {x_feature[feat2],y_feature[feat2],z_feature[feat2]};
							float xyzfeat_score2 = strength_feature[feat2];
							
							double feat_2dWire = -1;
							double feat_TimeT = -1;
							try
								{
								feat_2dWire = geom->NearestWire(xyzfeat2, plane, tpc, cstat);
								feat_TimeT  = detprop->ConvertXToTicks(x_feature[feat2], plane, tpc, cstat);
								}
							catch(...)
								{
								mf::LogWarning("FeatureVertexFinder") << "2dWire failed";
								continue;
								}
							//std::cout<<" Feat_2dWire = "<<feat_2dWire<<std::endl;
							//std::cout<<" Feat_timeT  = "<<feat_TimeT<<std::endl;
							// ######################################################################################
							// ###           Now check to see if this feature point is within 2 wires and         ###
							// ###   5 time ticks of the start time of the longest cluster found in this plane    ###
							// ######################################################################################
							if( std::abs(feat_2dWire - EndWire_LongestClusterInPlane) <= 2 && std::abs(EndTime_LongestClusterInPlane - feat_TimeT ) <=5 && 
							    xyzfeat_score2 >= temp_feat_score )
								{
								
								temp_feat_score = xyzfeat_score2;
								// ###########################################################
								// ### If a feature matches the 2d point increase strength ###
								// ###########################################################
								TwoDvertexStrength++;
								
								if(NothingFoundYet)
									{
									
									// ### Vertex found ###
									//std::cout<<std::endl;
									//std::cout<<" ##### Longest Cluster Feature Match Start Position #####"<<std::endl;
									//std::cout<<"X = "<<xyzfeat2[0]<<" , Y = "<<xyzfeat2[1]<<" , Z = "<<xyzfeat2[2]<<std::endl;
									//std::cout<<"Plane = "<<plane<<" , Wire = "<<feat_2dWire<<" , Time = "<<feat_TimeT<<std::endl;
									//std::cout<<std::endl;
								
									x_featClusMatch = xyzfeat2[0];
									y_featClusMatch = xyzfeat2[1];
									z_featClusMatch = xyzfeat2[2];
									
									double xyz[3] = {xyzfeat2[0], xyzfeat2[1], xyzfeat2[2]};
									double EndPoint2d_TimeTick = detprop->ConvertXToTicks(xyzfeat2[0],plane, tpc, cstat);
									
									try
										{
										int EndPoint2d_Wire 	   = geom->NearestWire(xyz , plane, tpc, cstat);
										int EndPoint2d_Channel     = geom->NearestChannel(xyz, plane, tpc, cstat);
										geo::View_t View	   = geom->View(EndPoint2d_Channel);
										geo::WireID wireID(cstat,tpc,plane,EndPoint2d_Wire);
									
										// ### Saving the 2d Vertex found ###
										recob::EndPoint2D vertex( EndPoint2d_TimeTick , //<---TimeTick
											  wireID ,		//<---geo::WireID
											  TwoDvertexStrength ,	//<---Vtx strength (JA: ?)
											  epcol->size() ,	//<---Vtx ID (JA: ?)
											  View ,		//<---Vtx View 	
											  1 );			//<---Vtx charge (JA: Need to figure this one?)
										epcol->push_back(vertex);
										}
									catch(...)
										{
										mf::LogWarning("FeatureVertexFinder") << "2dWire failed";
	      									continue;
										
										}
								
								
									}//<---End saving this vertex found
								
								
								NothingFoundYet = false;
								}//<---End finding a match between the 
							}//<---End feat for loop
					
						}//<---End Nothing Found Yet
					
					// ####################################################################################	
					// ### After all that...if we still don't have a 2-d / 3d Vertex then we just take  ###
					// ### the starting point of the longest cluster as out 2d vertex and in each plane ###
					// ### For the 3d point lets take the Feature with the greates strength as the 3-d  ###
					// ### and for now (JA: come back to) we won't enforce that these points match      ###
					// ####################################################################################
					if (NothingFoundYet)
						{
						
						double EndPoint2d_TimeTick = StartTime_LongestClusterInPlane;
						int EndPoint2d_Wire 	   = StartWire_LongestClusterInPlane;
						int EndPoint2d_Channel     = geom->PlaneWireToChannel(plane,EndPoint2d_Wire,tpc,cstat);
						geo::View_t View	   = geom->View(EndPoint2d_Channel);
						geo::WireID wireID(cstat,tpc,plane,EndPoint2d_Wire);
						
						// ### Saving the 2d Vertex found ###
						recob::EndPoint2D vertex( EndPoint2d_TimeTick , //<---TimeTick
									  wireID ,		//<---geo::WireID
									  TwoDvertexStrength ,	//<---Vtx strength (JA: ?)
									  epcol->size() ,	//<---Vtx ID (JA: ?)
									  View ,		//<---Vtx View 	
									  1 );			//<---Vtx Strength (JA: Need to figure this one?)
						epcol->push_back(vertex);
						
						
						
						
						
						}//<---End last check for something found
						
					LongestClusterStartWire[plane] = StartWire_LongestClusterInPlane;
					LongestClusterStartTime[plane] = StartTime_LongestClusterInPlane;
					}//<--End looping over planes
				}//<--End looping over TPC's
			}//<---End looping over cryostats
		
		// ########################################################
		// ### Last Ditch attempt to get a 3d point (if needed) ###'
		// ########################################################
		if (NothingFoundYet)
			{
			
			// ### Taking the highest scoring feature point ###
			float tempFeatureScore = 0;
			// ### Looping over features ###
			for (int feat3 = 0; feat3 < n3dFeatures; feat3++)
				{
							
				if( strength_feature[feat3]  >  tempFeatureScore)
					{
					tempFeatureScore = strength_feature[feat3];
					x_featClusMatch = x_feature[feat3];
					y_featClusMatch = y_feature[feat3];
					z_featClusMatch = z_feature[feat3];
								
								
								
					}//<---End checking if the current feature is the strongest feature
							
				}//<---End feat3 loop
			
			
			if (x_featClusMatch != 0) {NothingFoundYet = false;}
			
			}//<---Last ditch attempt to get a 3d vertex
			
		// ### In a last, last ditch effort we will take the start point
		// ### of the two longest clusters we found in each plane and just compute
		// ### an XYZ coordinate for them...regardless of how well they match
		if (NothingFoundYet)
			{
			double temp_wire_pitch0 = geom->WirePitch(0,1,0,0,0);
			double temp_wire_pitch1 = geom->WirePitch(0,1,1,0,0);
			double Iw0 = (LongestClusterStartWire[0]+3.95)*temp_wire_pitch0;
		        double Cw0 = (LongestClusterStartWire[0]+1.84)*temp_wire_pitch1;
			double It0 = LongestClusterStartTime[0] - presamplings;
			It0 *= TimetoCm;
			double Ct0 = LongestClusterStartTime[1] - presamplings ;
			Ct0 *= TimetoCm;
			x_featClusMatch = Ct0;
			y_featClusMatch = (Cw0-Iw0)/(2.*TMath::Sin(Angle));
			z_featClusMatch = (Cw0+Iw0)/(2.*TMath::Cos(Angle))-DetectorLength/2.*TMath::Tan(Angle);
			
			}
		
		double xyz2[3] = {x_featClusMatch, y_featClusMatch, z_featClusMatch};
			recob::Vertex the3Dvertex(xyz2, vcol->size());
			vcol->push_back(the3Dvertex);	

		}//<---End case where there was no good match	
       	
    
//----------------------------------------------------------------------------------------------------------------------------- 
    

    mf::LogVerbatim("Summary") << std::setfill('-') << std::setw(175) << "-" << std::setfill(' ');
    mf::LogVerbatim("Summary") << "FeatureVertexFinder Summary:";
    //for(size_t i = 0; i<epcol->size(); ++i) mf::LogVerbatim("Summary") << epcol->at(i) ;
    //for(size_t i = 0; i<vcol->size(); ++i) mf::LogVerbatim("Summary") << vcol->at(i) ;
    
    
    for(size_t j = 0; j<epcol->size(); ++j) std::cout<<" EndPoint2d = " << epcol->at(j) ;
    std::cout<<std::endl;
    for(size_t j = 0; j<vcol->size(); ++j) {std::cout<< " Vertex 3d = " << vcol->at(j) << std::endl;}
    std::cout<<std::endl;
    
    evt.put(std::move(epcol));
    evt.put(std::move(vcol));
    evt.put(std::move(assnep));
    evt.put(std::move(assntr));
    evt.put(std::move(assnsh));
    evt.put(std::move(assnh));

  } // end of produce
}


// =====================================================================================================
// =====================================================================================================






// =====================================================================================================
// =====================================================================================================
namespace vertex{

  DEFINE_ART_MODULE(FeatureVertexFinder)

} // end of vertex namespace
// =====================================================================================================
// =====================================================================================================

